#define _GNU_SOURCE

#include <time.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <fcntl.h>
#include <stdint.h>
#include <sys/syscall.h>
#include <string.h>

#define CLONE_INTO_CGROUP 0x200000000ULL /* Clone into a specific cgroup given the right permissions. */

int finished_children = 0;

void sighup()
{
    signal(SIGINT, sighup); /* reset signal */
    usleep(50000 - 12);
}

void proc_exit(int signum)
{
    finished_children += 1;
    if (finished_children == 9)
        exit(0);
}

struct weight {
    pid_t pid;
    int weight;
};

int dirfd_open_opath(const char *dir)
{
        return open(dir, O_RDONLY | O_PATH);
}
#define __aligned_u64 uint64_t __attribute__((aligned(8)))

struct __clone_args {
        __aligned_u64 flags;
        __aligned_u64 pidfd;
        __aligned_u64 child_tid;
        __aligned_u64 parent_tid;
        __aligned_u64 exit_signal;
        __aligned_u64 stack;
        __aligned_u64 stack_size;
        __aligned_u64 tls;
        __aligned_u64 set_tid;
        __aligned_u64 set_tid_size;
        __aligned_u64 cgroup;
};

pid_t clone_into_cgroup(int cgroup_fd)
{
        pid_t pid;
        struct __clone_args args = {
                .flags = CLONE_INTO_CGROUP,
                .exit_signal = SIGCHLD,
                .cgroup = cgroup_fd,
        };
    pid = syscall(SYS_clone3, &args, sizeof(struct __clone_args));

        if (pid < 0)
                goto pretend_enosys;

        return pid;

pretend_enosys:
        return -1;
}

int cmpfunc (const void * a, const void * b) {
   return ( ((struct weight*)a)->weight - ((struct weight*)b)->weight );
}

int main(int argc, char *argv[]) {
    int i, n = 9;
    pid_t pids[9];
    struct weight weights[9];
    char cgname[100] = "/sys/fs/cgroup/t0";
    int cgroup_fd;

    signal (SIGCHLD, proc_exit);

    for (i = 0; i < n; ++i) {
        // printf("%d\n", i);
        cgroup_fd = dirfd_open_opath(cgname);
        pids[i] = clone_into_cgroup(cgroup_fd);
        close(cgroup_fd);
        if (pids[i] < 0) {
            perror("fork");
            abort();
        } else if (pids[i] == 0) {
            printf("fork successfully %d\n", getppid());
            int t1 = 0, t2 = 1;
            int nextTerm = t1 + t2;
            struct timespec ts_start, ts_end;

            signal(SIGINT, sighup);
            timespec_get(&ts_start, TIME_UTC);
            for (int j = 0; j < 100000000000; j++) {
                t1 = t2;
                t2 = nextTerm;
                nextTerm = t1 + t2;
            }
            timespec_get(&ts_end, TIME_UTC);
            printf("latency label image: %d, %ld\n", i,  ts_end.tv_sec * 1000000 + (int)(ts_end.tv_nsec / 1e3) -
                                            ts_start.tv_sec * 1000000 - (int)(ts_start.tv_nsec / 1e3));
            exit(0);
        } else {
            printf("pids[%d]: %d\n", i, pids[i]);
        }
        // usleep(1000000);
    }
    printf("parent finishes clone children\n");
    usleep(5000);
    srand(getpid());
    while(1) {
        int rand_int = rand() % 100;
        int num_wakeup_tasks = -1;
        if (rand_int < 2) {
            num_wakeup_tasks = 0;
        } else if (rand_int < 8) {
            num_wakeup_tasks = 1;
        } else if (rand_int < 33) {
            num_wakeup_tasks = 2;
        } else if (rand_int < 40) {
            num_wakeup_tasks = 3;
        } else if (rand_int < 44) {
            num_wakeup_tasks = 4;
        } else if (rand_int < 50) {
            num_wakeup_tasks = 5;
        } else if (rand_int < 58) {
            num_wakeup_tasks = 6;
        } else if (rand_int < 90) {
            num_wakeup_tasks = 7;
        } else if (rand_int < 99) {
            num_wakeup_tasks = 8;
        } else{
            num_wakeup_tasks = 9;
        }
        // printf("num_wakeup_tasks: %d\n", num_wakeup_tasks);

        for (int j = 0; j < 9; j++) {
            int rand_weight = rand() % 100;
            weights[j].pid = pids[j];
            weights[j].weight = rand_weight;
        }
        qsort(weights, 9, sizeof(struct weight), cmpfunc);

        for (int i = 0; i < 9 - num_wakeup_tasks; i++) {
            kill(weights[i].pid, SIGINT);
            usleep(1);
        }

        usleep(50000);
    }

}
