#define _GNU_SOURCE

#include <time.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <fcntl.h>
#include <stdint.h>
#include <sys/syscall.h>
#include <string.h>

#define CLONE_INTO_CGROUP 0x200000000ULL /* Clone into a specific cgroup given the right permissions. */

int finished_children = 0;


void proc_exit(int signum)
{
    finished_children += 1;
    if (finished_children == 9)
        exit(0);
}

int dirfd_open_opath(const char *dir)
{
        return open(dir, O_RDONLY | O_PATH);
}
#define __aligned_u64 uint64_t __attribute__((aligned(8)))

struct __clone_args {
        __aligned_u64 flags;
        __aligned_u64 pidfd;
        __aligned_u64 child_tid;
        __aligned_u64 parent_tid;
        __aligned_u64 exit_signal;
        __aligned_u64 stack;
        __aligned_u64 stack_size;
        __aligned_u64 tls;
        __aligned_u64 set_tid;
        __aligned_u64 set_tid_size;
        __aligned_u64 cgroup;
};

pid_t clone_into_cgroup(int cgroup_fd)
{
        pid_t pid;
        struct __clone_args args = {
                .flags = CLONE_INTO_CGROUP,
                .exit_signal = SIGCHLD,
                .cgroup = cgroup_fd,
        };
        pid = syscall(SYS_clone3, &args, sizeof(struct __clone_args));

        if (pid < 0)
                goto pretend_enosys;

        return pid;

pretend_enosys:
        return -1;
}

int main(int argc, char *argv[]) {
    int i, n = 10;
    pid_t pids[10];
    char cgname[100] = "/sys/fs/cgroup/t1";
    int cgroup_fd;

    signal (SIGCHLD, proc_exit);

    for (i = 0; i < n; ++i) {
        // printf("%d\n", i);
        cgroup_fd = dirfd_open_opath(cgname);
        pids[i] = clone_into_cgroup(cgroup_fd);
        close(cgroup_fd);
        if (pids[i] < 0) {
            perror("fork");
            abort();
        } else if (pids[i] == 0) {
            printf("fork successfully %d\n", getppid());
            int t1 = 0, t2 = 1;
            int nextTerm = t1 + t2;
            struct timespec ts_start, ts_end;

            for (int j = 0; j < 100000000000; j++) {
                t1 = t2;
                t2 = nextTerm;
                nextTerm = t1 + t2;
            }
            exit(0);
        } else {
            printf("pids[%d]: %d\n", i, pids[i]);
        }
        // usleep(1000000);
    }
    printf("parent finishes clone children\n");
    usleep(5000);
    srand(getpid());
    while(1) {
        usleep(50000);
    }

}